@@ -86,7 +86,7 @@ from autobahn.xbr._blockchain import SimpleBlockchain
 from autobahn.xbr._buyer import SimpleBuyer
 from autobahn.xbr._seller import SimpleSeller, KeySeries
 from autobahn.xbr._interfaces import IMarketMaker, IProvider, IConsumer, ISeller, IBuyer
-from autobahn.xbr._util import sign_eip712_data, recover_eip712_signer
+from autobahn.xbr._util import sign_eip712_data, recover_eip712_signer, pack_uint256, unpack_uint256
 
 
 xbrtoken = None
@@ -164,6 +164,8 @@ __all__ = (
     'xbrnetwork',
     'sign_eip712_data',
     'recover_eip712_signer',
+    'pack_uint256',
+    'unpack_uint256',
 
     'MemberLevel',
     'ActorType',
